#
# Copyright 2017-2019 B-Open Solutions srl.
# Copyright 2017-2019 European Centre for Medium-Range Weather Forecasts (ECMWF).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# requires a Python 3 interpreter
import sys
if sys.version_info[0] < 3:
    raise EnvironmentError("Metview's Python interface requires Python 3. You are using Python "
                           + repr(sys.version_info))


# if the user has started via "python -m metview selfcheck"
# then we do not want to import anything yet because we want to
# catch errors differently

if len(sys.argv) != 2 or sys.argv[0] != "-m" or sys.argv[1] != "selfcheck":

    from . import bindings as _bindings

    _bindings.bind_functions(globals(), module_name=__name__)

    # Remove "_bindings" from the public API.
    del _bindings



"""
Metview Python use case

UC-04. The Analyst retrieves, for a given time interval, the values of
two parameters and combines their values on the same map

--------------------------------------------------------------------------------
1. Analyst retrieves, for a given time interval, the values of two chosen
   parameters (e.g. temperature, and geopotential) from a given source (i.e. MARS,
   files, observation databases)
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
2. Analyst customises many features of his map for each field he wants to plot
   (e.g. temperature field as shaded areas and geopotenti2. al field as isolines)
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
3. Analyst plots the data
--------------------------------------------------------------------------------
Analyst plots data variable t2 with contouring definition t_shade_c, and data
variable z with contouring definition mslp_isolines.
The fields will be plotted in the order they appear in the mv.plot() command,
with the shaded temperature at the bottom, and the geopotential on top.
"""

import metview as mv


# read 2m temperature
t2 = mv.read('./t2_for_UC-04.grib')

# read geopotential
z = mv.read('./z_for_UC-04.grib')

t_shade_c = mv.mcont(
    legend = True,
    contour_highlight = False,
    contour_level_selection_type = "interval",
    contour_interval = 10,
    contour_shade = True,
    contour_shade_max_level = 60,
    contour_shade_min_level = -60,
    contour_shade_method = "area_fill",
    contour_shade_max_level_colour = "red",
    contour_shade_min_level_colour = "blue",
    contour_shade_colour_direction = "clockwise"
 )

z_isolines = mv.mcont(
    legend = True,
    contour_line_thickness = 2,
    contour_line_colour = 'black',
    contour_highlight_colour = 'black',
    contour_highlight_thickness = 4,
    contour_level_selection_type = 'interval',
    contour_interval = 5,
    contour_legend_text = 'Geopotential',
)

mv.setoutput(mv.png_output(output_width = 1000, output_name = './gribplot'))
mv.plot(t2, t_shade_c, z, z_isolines)



#
# Copyright 2017-2019 B-Open Solutions srl.
# Copyright 2017-2019 European Centre for Medium-Range Weather Forecasts (ECMWF).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import keyword
import os
import pkgutil
import signal
import tempfile
import builtins
from enum import Enum

import cffi
import numpy as np


def string_from_ffi(s):
    return ffi.string(s).decode('utf-8')


# -----------------------------------------------------------------------------
#                                 Startup
# -----------------------------------------------------------------------------


class MetviewInvoker:
    """Starts a new Metview session on construction and terminates it on program exit"""

    def __init__(self):
        """
        Constructor - starts a Metview session and reads its environment information
        Raises an exception if Metview does not respond within 5 seconds
        """

        self.debug = (os.environ.get("METVIEW_PYTHON_DEBUG", '0') == '1')

        # check whether we're in a running Metview session
        if 'METVIEW_TITLE_PROD' in os.environ:
            self.persistent_session = True
            self.info_section = {'METVIEW_LIB': os.environ['METVIEW_LIB']}
            return

        import atexit
        import time
        import subprocess

        if self.debug:
            print('MetviewInvoker: Invoking Metview')
        self.persistent_session = False
        self.metview_replied = False
        self.metview_startup_timeout = 5  # seconds

        # start Metview with command-line parameters that will let it communicate back to us
        env_file = tempfile.NamedTemporaryFile(mode='rt')
        pid = os.getpid()
        # print('PYTHON:', pid, ' ', env_file.name, ' ', repr(signal.SIGUSR1))
        signal.signal(signal.SIGUSR1, self.signal_from_metview)
        # p = subprocess.Popen(['metview', '-edbg', 'tv8 -a', '-slog', '-python-serve',
        #     env_file.name, str(pid)], stdout=subprocess.PIPE)
        metview_startup_cmd = os.environ.get("METVIEW_PYTHON_START_CMD", 'metview')
        metview_flags = [metview_startup_cmd, '-nocreatehome', '-python-serve',
                         env_file.name, str(pid)]
        if self.debug:
            metview_flags.insert(2, '-slog')
            print('Starting Metview using these command args:')
            print(metview_flags)

        try:
            subprocess.Popen(metview_flags)
        except Exception as exp:
            print("Could not run the Metview executable ('" + metview_startup_cmd + "'); "
                  "check that the binaries for Metview (version 5 at least) are installed "
                  "and are in the PATH.")
            raise exp

        # wait for Metview to respond...
        wait_start = time.time()
        while (not(self.metview_replied) and
               (time.time() - wait_start < self.metview_startup_timeout)):
            time.sleep(0.001)

        if not(self.metview_replied):
            raise Exception('Command "metview" did not respond within '
                            + str(self.metview_startup_timeout) + ' seconds. '
                            'At least Metview 5 is required, so please ensure it is in your PATH, '
                            'as earlier versions will not work with the Python interface.')

        self.read_metview_settings(env_file.name)

        # when the Python session terminates, we should destroy this object so that the Metview
        # session is properly cleaned up. We can also do this in a __del__ function, but there can
        # be problems with the order of cleanup - e.g. the 'os' module might be deleted before
        # this destructor is called.
        atexit.register(self.destroy)

    def destroy(self):
        """Kills the Metview session. Raises an exception if it could not do it."""

        if self.persistent_session:
            return

        if self.metview_replied:
            if self.debug:
                print('MetviewInvoker: Closing Metview')
            metview_pid = self.info('EVENT_PID')
            try:
                os.kill(int(metview_pid), signal.SIGUSR1)
            except Exception as exp:
                print("Could not terminate the Metview process pid=" + metview_pid)
                raise exp

    def signal_from_metview(self, *args):
        """Called when Metview sends a signal back to Python to say that it's started"""
        # print ('PYTHON: GOT SIGNAL BACK FROM METVIEW!')
        self.metview_replied = True

    def read_metview_settings(self, settings_file):
        """Parses the settings file generated by Metview and sets the corresponding env vars"""
        import configparser

        cf = configparser.ConfigParser()
        cf.read(settings_file)
        env_section = cf['Environment']
        for envar in env_section:
            # print('set ', envar.upper(), ' = ', env_section[envar])
            os.environ[envar.upper()] = env_section[envar]
        self.info_section = cf['Info']

    def info(self, key):
        """Returns a piece of Metview information that was not set as an env var"""
        return self.info_section[key]

    def store_signal_handlers(self):
        """Stores the set of signal handlers that Metview will override"""
        self.sigint   = signal.getsignal(signal.SIGINT)
        self.sighup   = signal.getsignal(signal.SIGHUP)
        self.sighquit = signal.getsignal(signal.SIGQUIT)
        self.sigterm  = signal.getsignal(signal.SIGTERM)
        self.sigalarm = signal.getsignal(signal.SIGALRM)

    def restore_signal_handlers(self):
        """Restores the set of signal handlers that Metview has overridden"""
        signal.signal(signal.SIGINT,  self.sigint)
        signal.signal(signal.SIGHUP,  self.sighup)
        signal.signal(signal.SIGQUIT, self.sighquit)
        signal.signal(signal.SIGTERM, self.sigterm)
        signal.signal(signal.SIGALRM, self.sigalarm)


mi = MetviewInvoker()

try:
    ffi = cffi.FFI()
    ffi.cdef(pkgutil.get_data('metview', 'metview.h').decode('ascii'))
    mv_lib = mi.info('METVIEW_LIB')
    # is there a more general way to add to a path to a list of paths?
    os.environ["LD_LIBRARY_PATH"] = mv_lib + ':' + os.environ.get("LD_LIBRARY_PATH", '')

    try:
        # Linux / Unix systems
        lib = ffi.dlopen(os.path.join(mv_lib, 'libMvMacro.so'))
    except OSError:
        # MacOS systems
        lib = ffi.dlopen(os.path.join(mv_lib, 'libMvMacro'))

except Exception as exp:
    print('Error loading Metview/libMvMacro. LD_LIBRARY_PATH='
          + os.environ.get("LD_LIBRARY_PATH", ''))
    raise exp


# The C/C++ code behind lib.p_init() will call marsinit(), which overrides various signal
# handlers. We don't necessarily want this when running a Python script - we should use
# the default Python behaviour for handling signals, so we save the current signals
# before calling p_init() and restore them after.
mi.store_signal_handlers()
lib.p_init()
mi.restore_signal_handlers()


# -----------------------------------------------------------------------------
#                        Classes to handle complex Macro types
# -----------------------------------------------------------------------------


class Value:

    def __init__(self, val_pointer):
        self.val_pointer = val_pointer

    def push(self):
        if self.val_pointer is None:
            lib.p_push_nil()
        else:
            lib.p_push_value(self.val_pointer)

    # if we steal a value pointer from a temporary Value object, we need to
    # ensure that the Metview Value is not destroyed when the temporary object
    # is destroyed by setting its pointer to None
    def steal_val_pointer(self, other):
        self.val_pointer = other.val_pointer
        other.val_pointer = None

    # enable a more object-oriented interface, e.g. a = fs.interpolate(10, 29.4)
    def __getattr__(self, fname):
        def call_func_with_self(*args, **kwargs):
            return call(fname, self, *args, **kwargs)
        return call_func_with_self

    # on destruction, ensure that the Macro Value is also destroyed
    def __del__(self):
        try:
            if self.val_pointer is not None and lib is not None:
                lib.p_destroy_value(self.val_pointer)
                self.val_pointer = None
        except Exception as exp:
            print("Could not destroy Metview variable ", self)
            raise exp


class Request(dict, Value):
    verb = "UNKNOWN"

    def __init__(self, req):
        self.val_pointer = None

        # initialise from Python object (dict/Request)
        if isinstance(req, dict):
            self.update(req)
            self.to_metview_style()
            if isinstance(req, Request):
                self.verb = req.verb
                self.val_pointer = req.val_pointer

        # initialise from a Macro pointer
        else:
            Value.__init__(self, req)
            self.verb = string_from_ffi(lib.p_get_req_verb(req))
            n = lib.p_get_req_num_params(req)
            for i in range(0, n):
                param = string_from_ffi(lib.p_get_req_param(req, i))
                raw_val = lib.p_get_req_value(req, param.encode('utf-8'))
                if raw_val != ffi.NULL:
                    val = string_from_ffi(raw_val)
                    self[param] = val
            # self['_MACRO'] = 'BLANK'
            # self['_PATH']  = 'BLANK'

    def __str__(self):
        return "VERB: " + self.verb + super().__str__()

    # translate Python classes into Metview ones where needed
    def to_metview_style(self):
        for k, v in self.items():

            # bool -> on/off
            if isinstance(v, bool):
                conversion_dict = {True: 'on', False: 'off'}
                self[k] = conversion_dict[v]

            # class_ -> class (because 'class' is a Python keyword and cannot be
            # used as a named parameter)
            elif k == 'class_':
                self['class'] = v
                del self['class_']

    def push(self):
        # if we have a pointer to a Metview Value, then use that because it's more
        # complete than the dict
        if self.val_pointer:
            Value.push(self)
        else:
            r = lib.p_new_request(self.verb.encode('utf-8'))

            # to populate a request on the Macro side, we push each
            # value onto its stack, and then tell it to create a new
            # parameter with that name for the request. This allows us to
            # use Macro to handle the addition of complex data types to
            # a request
            for k, v in self.items():
                push_arg(v)
                lib.p_set_request_value_from_pop(r, k.encode('utf-8'))

            lib.p_push_request(r)

    def __getitem__(self, index):
        return subset(self, index)


def push_bytes(b):
    lib.p_push_string(b)


def push_str(s):
    push_bytes(s.encode('utf-8'))


def push_list(lst):
    # ask Metview to create a new list, then add each element by
    # pusing it onto the stack and asking Metview to pop it off
    # and add it to the list
    mlist = lib.p_new_list(len(lst))
    for i, val in enumerate(lst):
        push_arg(val)
        lib.p_add_value_from_pop_to_list(mlist, i)
    lib.p_push_list(mlist)


def push_date(d):
    lib.p_push_datestring(np.datetime_as_string(d).encode('utf-8'))


def push_datetime(d):
    lib.p_push_datestring(d.isoformat().encode('utf-8'))


def push_datetime_date(d):
    s = d.isoformat() + 'T00:00:00'
    lib.p_push_datestring(s.encode('utf-8'))


def push_vector(npa):

    # convert numpy array to CData
    if npa.dtype == np.float64:
        cffi_buffer = ffi.cast('double*', npa.ctypes.data)
        lib.p_push_vector_from_double_array(cffi_buffer, len(npa), np.nan)
    elif npa.dtype == np.float32:
        cffi_buffer = ffi.cast('float*', npa.ctypes.data)
        lib.p_push_vector_from_float32_array(cffi_buffer, len(npa), np.nan)
    else:
        raise Exception('Only float32 and float64 numPy arrays can be passed to Metview, not ',
                        npa.dtype)


class FileBackedValue(Value):

    def __init__(self, val_pointer):
        Value.__init__(self, val_pointer)

    def url(self):
        # ask Metview for the file relating to this data (Metview will write it if necessary)
        return string_from_ffi(lib.p_data_path(self.val_pointer))


class FileBackedValueWithOperators(FileBackedValue):

    def __init__(self, val_pointer):
        FileBackedValue.__init__(self, val_pointer)

    def __add__(self, other):
        return add(self, other)

    def __sub__(self, other):
        return sub(self, other)

    def __mul__(self, other):
        return prod(self, other)

    def __truediv__(self, other):
        return div(self, other)

    def __pow__(self, other):
        return power(self, other)

    def __ge__(self, other):
        return greater_equal_than(self, other)

    def __gt__(self, other):
        return greater_than(self, other)

    def __le__(self, other):
        return lower_equal_than(self, other)

    def __lt__(self, other):
        return lower_than(self, other)

    def __eq__(self, other):
        return equal(self, other)

    def __ne__(self, other):
        return met_not_eq(self, other)


class ContainerValue(Value):
    def __init__(self, val_pointer, macro_index_base, element_type, support_slicing):
        Value.__init__(self, val_pointer)
        self.idx = 0
        self.macro_index_base = macro_index_base
        self.element_type = element_type  # the type of elements that the container contains
        self.support_slicing = support_slicing

    def __len__(self):
        if self.val_pointer is None:
            return 0
        else:
            return int(count(self))

    def __getitem__(self, index):
        if isinstance(index, slice):
            if self.support_slicing:
                indices = index.indices(len(self))
                fields = [self[i] for i in range(*indices)]
                if len(fields) == 0:
                    return None
                else:
                    f = fields[0]
                    for i in range(1, len(fields)):
                        f = merge(f, fields[i])
                    return f
            else:
                raise Exception('This object does not support extended slicing: ' + str(self))
        else:  # normal index
            if isinstance(index, str):  # can have a string as an index
                return subset(self, index)
            else:
                return subset(self, index + self.macro_index_base)  # numeric index: 0->1-based

    def __setitem__(self, index, value):
        if (isinstance(value, self.element_type)):
            lib.p_set_subvalue(self.val_pointer, index + self.macro_index_base, value.val_pointer)
        else:
            raise Exception('Cannot assign ', value, ' as element of ', self)

    def __iter__(self):
        return self

    def __next__(self):
        if self.idx >= self.__len__():
            self.idx = 0
            raise StopIteration
        else:
            self.idx += 1
            return self.__getitem__(self.idx - 1)


class Fieldset(FileBackedValueWithOperators, ContainerValue):

    def __init__(self, val_pointer=None, path=None):
        FileBackedValue.__init__(self, val_pointer)
        ContainerValue.__init__(self, val_pointer, 1, Fieldset, True)
        if path is not None:
            temp = read(path)
            self.steal_val_pointer(temp)

    def append(self, other):
        temp = merge(self, other)
        self.steal_val_pointer(temp)

    def to_dataset(self):
        # soft dependency on cfgrib
        try:
            from cfgrib import xarray_store
        except ImportError:
            print("Package cfgrib/xarray_store not found. Try running 'pip install cfgrib'.")
            raise
        dataset = xarray_store.open_dataset(self.url())
        return dataset


class Bufr(FileBackedValue):

    def __init__(self, val_pointer):
        FileBackedValue.__init__(self, val_pointer)


class Geopoints(FileBackedValueWithOperators, ContainerValue):

    def __init__(self, val_pointer):
        FileBackedValueWithOperators.__init__(self, val_pointer)
        ContainerValue.__init__(self, val_pointer, 0, None, False)

    def to_dataframe(self):
        try:
            import pandas as pd
        except ImportError:
            print("Package pandas not found. Try running 'pip install pandas'.")
            raise

        # create a dictionary of columns (note that we do not include 'time'
        # because it is incorporated into 'date')
        cols = self.columns()
        if 'time' in cols:
            cols.remove('time')

        pddict = {}
        for c in cols:
            pddict[c] = self[c]

        df = pd.DataFrame(pddict)
        return df


class NetCDF(FileBackedValueWithOperators):
    def __init__(self, val_pointer):
        FileBackedValueWithOperators.__init__(self, val_pointer)

    def to_dataset(self):
        # soft dependency on xarray
        try:
            import xarray as xr
        except ImportError:
            print("Package xarray not found. Try running 'pip install xarray'.")
            raise
        dataset = xr.open_dataset(self.url())
        return dataset


class Odb(FileBackedValue):

    def __init__(self, val_pointer):
        FileBackedValue.__init__(self, val_pointer)

    def to_dataframe(self):
        try:
            import pandas as pd
        except ImportError:
            print("Package pandas not found. Try running 'pip install pandas'.")
            raise

        cols = self.columns()
        pddict = {}

        for col in cols:
            pddict[col] = self.values(col)

        df = pd.DataFrame(pddict)
        return df


class Table(FileBackedValue):

    def __init__(self, val_pointer):
        FileBackedValue.__init__(self, val_pointer)

    def to_dataframe(self):
        try:
            import pandas as pd
        except ImportError:
            print("Package pandas not found. Try running 'pip install pandas'.")
            raise

        df = pd.read_csv(self.url())
        return df


class GeopointSet(FileBackedValueWithOperators, ContainerValue):

    def __init__(self, val_pointer):
        FileBackedValueWithOperators.__init__(self, val_pointer)
        ContainerValue.__init__(self, val_pointer, 1, Geopoints, False)


# -----------------------------------------------------------------------------
#                        Pushing data types to Macro
# -----------------------------------------------------------------------------


def dataset_to_fieldset(val, **kwarg):
    # we try to import xarray as locally as possible to reduce startup time
    # try to write the xarray as a GRIB file, then read into a fieldset
    import xarray as xr
    import cfgrib

    if not isinstance(val, xr.core.dataset.Dataset):
        raise TypeError('dataset_to_fieldset requires a variable of type xr.core.dataset.Dataset;'
                        ' was supplied with ', builtins.type(val))

    f, tmp = tempfile.mkstemp(".grib")
    os.close(f)

    try:
        # could add keys, e.g. grib_keys={'centre': 'ecmf'})
        cfgrib.to_grib(val, tmp, **kwarg)
    except:
        print("Error trying to write xarray dataset to GRIB for conversion to Metview Fieldset")
        raise

    # TODO: tell Metview that this is a temporary file that should be deleted when no longer needed
    fs = read(tmp)
    return fs


def push_xarray_dataset(val):
    fs = dataset_to_fieldset(val)
    fs.push()


# try_to_push_complex_type exists as a separate function so that we don't have
# to import xarray at the top of the module - this saves some time on startup
def try_to_push_complex_type(val):
    import xarray as xr
    if isinstance(val, xr.core.dataset.Dataset):
        push_xarray_dataset(val)
    else:
        raise TypeError('Cannot push this type of argument to Metview: ', builtins.type(val))


class ValuePusher():
    """Class to handle pushing values to the Macro library"""

    def __init__(self):
        # a set of pairs linking value types with functions to push them to Macro
        # note that Request must come before dict, because a Request inherits from dict;
        # this ordering requirement also means we should use list or tuple instead of a dict
        self.funcs = (
            (float,             lambda n : lib.p_push_number(n)),
            ((int, np.number),  lambda n : lib.p_push_number(float(n))),
            (str,               lambda n : push_str(n)),
            (Request,           lambda n : n.push()),
            (dict,              lambda n : Request(n).push()),
            ((list, tuple),     lambda n : push_list(n)),
            (type(None),        lambda n : lib.p_push_nil()),
            (FileBackedValue,   lambda n : n.push()),
            (np.datetime64,     lambda n : push_date(n)),
            (datetime.datetime, lambda n : push_datetime(n)),
            (datetime.date,     lambda n : push_datetime_date(n)),
            (np.ndarray,        lambda n : push_vector(n)),
        )

    def push_value(self, val):
        for typekey, typefunc in self.funcs:
            if isinstance(val, typekey):
                typefunc(val)
                return 1

        # if we haven't returned yet, then try the more complex types
        try_to_push_complex_type(val)
        return 1

vp = ValuePusher()


def push_arg(n):
    return vp.push_value(n)


def dict_to_pushed_args(d):

    # push each key and value onto the argument stack
    for k, v in d.items():
        push_str(k)
        push_arg(v)

    return 2 * len(d)  # return the number of arguments generated


# -----------------------------------------------------------------------------
#                        Returning data types from Macro
# -----------------------------------------------------------------------------


def list_from_metview(val):

    mlist = lib.p_value_as_list(val)
    result = []
    n = lib.p_list_count(mlist)
    all_vectors = True
    for i in range(0, n):
        mval = lib.p_list_element_as_value(mlist, i)
        v = value_from_metview(mval)
        if all_vectors and not isinstance(v, np.ndarray):
            all_vectors = False
        result.append(v)

    # if this is a list of vectors, then create a 2-D numPy array
    if all_vectors and n > 0:
        result = np.stack(result, axis=0)

    return result


def datestring_from_metview(val):

    mdate = string_from_ffi(lib.p_value_as_datestring(val))
    dt = datetime.datetime.strptime(mdate, "%Y-%m-%dT%H:%M:%S")
    return dt


def vector_from_metview(val):

    vec = lib.p_value_as_vector(val, np.nan)

    n = lib.p_vector_count(vec)
    s = lib.p_vector_elem_size(vec)

    if s == 4:
        nptype = np.float32
        b = lib.p_vector_float32_array(vec)
    elif s == 8:
        nptype = np.float64
        b = lib.p_vector_double_array(vec)
    else:
        raise Exception('Metview vector data type cannot be handled: ', s)

    bsize = n * s
    c_buffer = ffi.buffer(b, bsize)
    np_array = np.frombuffer(c_buffer, dtype=nptype)
    return np_array


def handle_error(val):
    msg = string_from_ffi(lib.p_error_message(val))
    if "Service" in msg and "Examiner" in msg:
        return None
    else:
        return Exception('Metview error: ' + (msg))


def string_from_metview(val):
    return string_from_ffi(lib.p_value_as_string(val))


class MvRetVal(Enum):
    tnumber  = 0
    tstring  = 1
    tgrib    = 2
    trequest = 3
    tbufr    = 4
    tgeopts  = 5
    tlist    = 6
    tnetcdf  = 7
    tnil     = 8
    terror   = 9
    tdate    = 10
    tvector  = 11
    todb     = 12
    ttable   = 13
    tgptset  = 14
    tunknown = 99


class ValueReturner():
    """Class to handle return values from the Macro library"""
    def __init__(self):
        self.funcs = {}
        self.funcs[MvRetVal.tnumber.value]  = lambda val : lib.p_value_as_number(val)
        self.funcs[MvRetVal.tstring.value]  = lambda val : string_from_metview(val)
        self.funcs[MvRetVal.tgrib.value]    = lambda val : Fieldset(val)
        self.funcs[MvRetVal.trequest.value] = lambda val : Request(val)
        self.funcs[MvRetVal.tbufr.value]    = lambda val : Bufr(val)
        self.funcs[MvRetVal.tgeopts.value]  = lambda val : Geopoints(val)
        self.funcs[MvRetVal.tlist.value]    = lambda val : list_from_metview(val)
        self.funcs[MvRetVal.tnetcdf.value]  = lambda val : NetCDF(val)
        self.funcs[MvRetVal.tnil.value]     = lambda val : None
        self.funcs[MvRetVal.terror.value]   = lambda val : handle_error(val)
        self.funcs[MvRetVal.tdate.value]    = lambda val : datestring_from_metview(val)
        self.funcs[MvRetVal.tvector.value]  = lambda val : vector_from_metview(val)
        self.funcs[MvRetVal.todb.value]     = lambda val : Odb(val)
        self.funcs[MvRetVal.ttable.value]   = lambda val : Table(val)
        self.funcs[MvRetVal.tgptset.value]  = lambda val : GeopointSet(val)

    def translate_return_val(self, val):
        rt = lib.p_value_type(val)
        try:
            return self.funcs[rt](val)
        except Exception:
            raise Exception('value_from_metview got an unhandled return type: ' + str(rt))


vr = ValueReturner()


def value_from_metview(val):
    retval = vr.translate_return_val(val)
    if isinstance(retval, Exception):
        raise retval
    return retval


# -----------------------------------------------------------------------------
#                        Creating and calling Macro functions
# -----------------------------------------------------------------------------

def _call_function(mfname, *args, **kwargs):

    nargs = 0

    for n in args:
        actual_n_args = push_arg(n)
        nargs += actual_n_args

    merged_dict = {}
    merged_dict.update(kwargs)
    if len(merged_dict) > 0:
        dn = dict_to_pushed_args(Request(merged_dict))
        nargs += dn

    lib.p_call_function(mfname.encode('utf-8'), nargs)


def make(mfname):

    def wrapped(*args, **kwargs):
        err = _call_function(mfname, *args, **kwargs)
        if err:
            pass  # throw Exceception

        val = lib.p_result_as_value()
        return value_from_metview(val)

    return wrapped


def bind_functions(namespace, module_name=None):
    """Add to the module globals all metview functions except operators like: +, &, etc."""
    for metview_name in make('dictionary')():
        if metview_name.isidentifier():
            python_name = metview_name
            # NOTE: we append a '_' to metview functions that clash with python reserved keywords
            #   as they cannot be used as identifiers, for example: 'in' -> 'in_'
            if keyword.iskeyword(metview_name):
                python_name += '_'
            python_func = make(metview_name)
            python_func.__name__ = python_name
            python_func.__qualname__ = python_name
            if module_name:
                python_func.__module__ = module_name
            namespace[python_name] = python_func
        # else:
        #    print('metview function %r not bound to python' % metview_name)
    # add the 'mvl' functions, which are written in Macro and therefore not
    # listed by the dictionary() function
    for f in ['mvl_ml2hPa', 'mvl_create_netcdf_2d', 'mvl_flextra_etadot', 'mvl_geocircle',
              'mvl_geoline', 'mvl_geopotential_on_ml', 'mvl_mxn_subframes', 'mvl_plot_scm_data',
              'mvl_regular_layout', 'mvl_regular_layout_area', 'thermo_data_info',
              'thermo_parcel_path', 'thermo_parcel_area', 'xy_curve', 'potential_temperature',
              'temperature_from_potential_temperature', 'saturation_mixing_ratio', 'mixing_ratio',
              'vapour_pressure', 'saturation_vapour_pressure',
              'lifted_condensation_level', 'divergence', 'vorticity', 'laplacian',
              'geostrophic_wind_pl', 'geostrophic_wind_ml']:
        namespace[f] = make(f)

    # HACK: some fuctions are missing from the 'dictionary' call.
    namespace['neg'] = make('neg')
    namespace['nil'] = make('nil')
    # override some functions that need special treatment
    # FIXME: this needs to be more structured
    namespace['plot'] = plot
    namespace['setoutput'] = setoutput
    namespace['dataset_to_fieldset'] = dataset_to_fieldset

    namespace['Fieldset'] = Fieldset


# some explicit bindings are used here
add = make('+')
call = make('call')
count = make('count')
div = make('/')
equal = make('=')
filter = make('filter')
greater_equal_than = make('>=')
greater_than = make('>')
lower_equal_than = make('<=')
lower_than = make('<')
merge = make('&')
met_not_eq = make('<>')
met_plot = make('plot')
nil = make('nil')
png_output = make('png_output')
power = make('^')
prod = make('*')
ps_output = make('ps_output')
read = make('read')
met_setoutput = make('setoutput')
sub = make('-')
subset = make('[]')


# -----------------------------------------------------------------------------
#                        Particular code for calling the plot() command
# -----------------------------------------------------------------------------

class Plot():

    def __init__(self):
        self.plot_to_jupyter = False

    def __call__(self, *args, **kwargs):
        if self.plot_to_jupyter:
            f, tmp = tempfile.mkstemp(".png")
            os.close(f)

            base, ext = os.path.splitext(tmp)

            met_setoutput(png_output(output_name=base, output_name_first_page_number='off'))
            met_plot(*args)

            image = Image(tmp)
            os.unlink(tmp)
            return image
        else:
            map_outputs = {
                'png': png_output,
                'ps': ps_output,
            }
            if 'output_type' in kwargs:
                output_function = map_outputs[kwargs['output_type'].lower()]
                kwargs.pop('output_type')
                met_plot(output_function(kwargs), *args)
            else:
                met_plot(*args)
            # the Macro plot command returns an empty definition, but
            # None is better for Python
            return None


plot = Plot()


# On a test system, importing IPython took approx 0.5 seconds, so to avoid that hit
# under most circumstances, we only import it when the user asks for Jupyter
# functionality. Since this occurs within a function, we need a little trickery to
# get the IPython functions into the global namespace so that the plot object can use them
def setoutput(*args):
    if 'jupyter' in args:
        try:
            global Image
            global get_ipython
            IPython = __import__('IPython', globals(), locals())
            Image = IPython.display.Image
            get_ipython = IPython.get_ipython
        except ImportError as imperr:
            print('Could not import IPython module - plotting to Jupyter will not work')
            raise imperr

        # test whether we're in the Jupyter environment
        if get_ipython() is not None:
            plot.plot_to_jupyter = True
        else:
            print("ERROR: setoutput('jupyter') was set, but we are not in a Jupyter environment")
            raise(Exception('Could not set output to jupyter'))
    else:
        plot.plot_to_jupyter = False
        met_setoutput(*args)



# ==============================================================================
# Authors: ralf mueller, stephan siemen
#
#
# Plan is to create a plot similar to the scatter plot for co2 concentration and
# september minimum of sea ice extent
#
# reference:
#   https://www.mpg.de/10579957/W004_Environment_climate_062-069.pdf, p. 7
#
# ==============================================================================
import os
from ecmwfapi import ECMWFDataServer
from cdo import Cdo
from multiprocessing import Pool
from tarfile import TarFile
import matplotlib.pyplot as plt
import matplotlib.transforms as mtrans
# basic setup    {{{ ===========================================================
server = ECMWFDataServer()
cdo    = Cdo()
cdo.debug = True
tasks     = 4
startYear = 1980
endYear   = 2014
# }}} ==========================================================================
# helper methods {{{ ===========================================================
def getDataFromTarfile(tarfile):
    tf = TarFile(tarfile)
    members = [ m.name for m in tf.getmembers()]   
    if (list(set([os.path.exists(x) for x in members])) != [True]):
        tf.extractall()
    tf.close
    return members

def computeTimeSeries(file,varname,useCellArea=False):
    if (useCellArea):
        ofile = cdo.mul(input = '-fldmean -selname,%s %s -fldsum -gridarea %s'%(varname,file,file),
                options = '-b F32',output = '_'+os.path.basename(file),force=False)
    else:
        ofile = cdo.fldmean(input =  '-selname,%s %s'%(varname,file),
                options = '-b F32',output = '_'+os.path.basename(file),force=False)
    return ofile

def computeTimeSeriesOfFilelist(pool,files,varname,ofile,useCellArea=False):
    results = dict()
    for file in files:
        rfile = pool.apply_async(computeTimeSeries,(file,varname,False))
        results[file] = rfile
    pool.close()
    pool.join()

    for k,v in results.items():
        results[k] = v.get()

    cdo.yearmean(input = '-cat %s'%(' '.join([results[x] for x in files])),
            output = ofile, force=False,
            options = '-f nc')
    return ofile
# }}} ==========================================================================
# Sea Ice Cover retrival + processing {{{
iceCover_file = "ci_interim_%s-%s-NH.grb"%(startYear, endYear)
if ( not os.path.exists(iceCover_file) ):
    server.retrieve({
        'stream'    : "oper",
        'levtype'   : "sfc",
        'param'     : "31.128",
        'dataset'   : "interim",
        'step'      : "0",
        'grid'      : "0.5/0.5",
        'time'      : "12",
        'date'      : "%s-01-01/to/%s-01-01"%(startYear,endYear),
        'type'      : "an",
        'class'     : "ei",
        'area'      : "90/-180/0/180",
        'target'    : iceCover_file
     })
else:
    print("use existing file '%s'"%(iceCover_file))
# compute the nh ice extent: minimum usually happens in September
iceExtent = 'ice_extent_%s-%s-daymean-SeptMin.nc'%(startYear,endYear)
cdo.setattribute('sea_ice_extent@unit=m2,sea_ice_extent@standard_name=sea_ice_extent',
         input  = '-setname,sea_ice_extent -yearmin -fldsum -mul -selmon,9 %s -gridarea %s'%(iceCover_file,iceCover_file),
         output = iceExtent,force=False,
         options = '-f nc')
iceExtent_ds = cdo.readXDataset(iceExtent)
# }}} ==========================================================================
# {{{ CO2 retrieval + processing ===========================================================
# cams return tarballs of netcdf files
co2_tarball = "co2_totalColumn_%s-%s.tar"%(startYear, endYear)
if ( not os.path.exists(co2_tarball) ):
    server.retrieve({ #CO2
        "dataset"   : "cams_ghg_inversions",
        "datatype"  : "ra",
        "date"      : "%s-01-01/to/%s-01-01"%(startYear,endYear),
        "frequency" : "3h",
        "param"     : "co2",
        "quantity"  : "total_column",
        "version"   : "v16r2",
        "target"    : co2_tarball
    })
else:
    print("use existing file '%s'"%(co2_tarball))
co2_files      = getDataFromTarfile(co2_tarball)
co2_timeSeries = 'co2_timeseries_%s-%s.nc'%(startYear,endYear)
computeTimeSeriesOfFilelist(Pool(tasks),co2_files,'XCO2',co2_timeSeries,False)
co2_ds = cdo.readXDataset(co2_timeSeries)

# }}} ==========================================================================
# scatter plot {{{ =============================================================
# some debugging output
iceExtent_ds.info()
co2_ds.info()
# shaping the data for plotting it
xSelection =       co2_ds.sel(time=slice('%s-01-01'%(startYear), '%s-01-01'%(endYear)))
ySelection = iceExtent_ds.sel(time=slice('%s-01-01'%(startYear), '%s-01-01'%(endYear)))
# create the final scatter plot
fig = plt.figure(figsize=(10, 7))
ax  = plt.subplot(1, 1, 1)
trans_offset = mtrans.offset_copy(ax.transData, fig=fig,
                                  x=0.05, y=-0.20, units='inches') # inches because we are in UK

x = xSelection.to_array()[1,:,0,0,0]
y = ySelection.to_array()[1,:,0,0,0]
plt.scatter( x , y)
# put years as labels
years = xSelection.time.dt.year
for _x,_y,_year in zip(x,y,years):
    plt.text(_x, _y, '%d'%(_year), transform=trans_offset)

plt.grid(True)
plt.ylabel('sea ice extent [m2]')
plt.xlabel('co2 concentration [ppm]')
plt.title('Correlation of NH Sea Ice extent minimum and CO2 emissions')
plt.savefig('seaIce_CO2_correlation.png')
# }}} ==========================================================================
# vim:fdm=marker




import os
import numpy as np
import datetime
import pytest
import pandas as pd
import xarray as xr

import metview as mv
from metview import bindings


PATH = os.path.dirname(__file__)
MAX_VALUE = 316.06060791015625
SEMI_EQUATOR = 20001600.0
MAX_SQRT_GPT = 16.867127793433
MAX_GPT = 284.5


def metview_version():
    version = mv.version_info()
    return version['metview_version']


def supports_float32_vectors():
    return metview_version() >= 50200


def test_names():
    assert mv.dictionary.__name__ == mv.dictionary.__qualname__ == 'dictionary'
    assert mv.dictionary.__module__ == 'metview'


def file_in_testdir(filename):
    return os.path.join(PATH, filename)


def test_version_info():
    out = mv.version_info()
    assert 'metview_version' in out


def test_describe():
    mv.describe('type')


def test_definitions():
    mcont_def = mv.mcont({'legend': True})
    msymb_def = mv.msymb({'symbol_type': 'marker'})
    mcoast_def = mv.mcoast({'map_coastline_land_shade': True})
    mobs_def = mv.mobs({'obs_temperature': False})
    mtext_def = mv.mtext({'text_font_size': '0.80'})
    geoview_def = mv.geoview({'map_projection': 'polar_stereographic'})
    ps_output_def = mv.ps_output({'output_name': 'test'})
    assert mcont_def['LEGEND'] == 'ON'
    assert msymb_def['SYMBOL_TYPE'] == 'MARKER'
    assert mcoast_def['MAP_COASTLINE_LAND_SHADE'] == 'ON'
    assert mobs_def['OBS_TEMPERATURE'] == 'OFF'
    assert mtext_def['TEXT_FONT_SIZE'] == 0.8
    assert geoview_def['MAP_PROJECTION'] == 'POLAR_STEREOGRAPHIC'
    assert ps_output_def['OUTPUT_NAME'] == 'test'


def test_print():
    mv.print('Start ', 7, 1, 3, ' Finished!')
    mv.print(6, 2, ' Middle ', 6)


def test_lowercase():
    a = mv.lowercase('MetViEw')
    assert a == 'metview'


# def test_lists():
#     m= mv.mcont(contour_level_selection_type = 'level_list', contour_level_list = [1, 2, 6])
#     print('M: ', m)
# test_lists()


def test_create_list():
    inlist = [1, 5, 6, 5, 1, 9, 18]
    outlist = mv.list(*inlist)
    assert outlist == inlist


def test_create_empty_list():
    outlist = mv.list()
    assert outlist == []


def test_create_list_from_tuple():
    intuple = (10, 50, 60, 50, 1.1, 90)
    outlist = mv.list(*intuple)
    assert outlist == list(intuple)


def test_list_unique():
    inlist = [1, 5, 6, 5, 1, 9]
    ulist = mv.unique(inlist)
    assert ulist == [1, 5, 6, 9]


def test_list_sort():
    s = [4, 7, 2, 3, 4, 9, 0]
    so = mv.sort(s)
    assert(so == [0, 2, 3, 4, 4, 7, 9])


def test_list_sort_indices():
    s = [4, 7, 2, 3, 4, 9, 0]
    soi = mv.sort_indices(s)
    assert(soi == [6, 2, 3, 0, 4, 1, 5])


def test_list_find():
    s = [4, 7, 2, 3, 4, 9, 0]
    f = mv.find(s, 2)
    assert(f == 2)
    f = mv.find(s, 4)
    assert(f == 0)
    f = mv.find(s, 4, 'all')
    assert(f == [0, 4])


def test_tuple_unique():
    intuple = (3, 2, 2, 7, 3, 1.2, 2.1, 1.2)
    ulist = mv.unique(intuple)
    assert ulist == [3, 2, 7, 1.2, 2.1]


def test_read():
    gg = mv.read({'SOURCE': file_in_testdir('test.grib'), 'GRID': 80})
    assert mv.grib_get_string(gg, 'typeOfGrid') == 'regular_gg'


def test_read_filter_1():
    mls = mv.read(source=file_in_testdir('ml_data.grib'), param='t', levelist=[5, 49, 101])
    assert(mv.grib_get_long(mls, 'level') == [5, 49, 101])


def test_read_filter_2():
    tuv_all_levs = mv.read(file_in_testdir('tuv_pl.grib'))
    u_all_levs = mv.read(data=tuv_all_levs, param='u')
    assert(mv.grib_get_long(u_all_levs, 'level') == [1000, 850, 700, 500, 400, 300])
    assert(mv.grib_get_string(u_all_levs, 'shortName') == ['u', 'u', 'u', 'u', 'u', 'u'])
    u_2levs = mv.read(data=u_all_levs, levelist=[700, 400])
    assert(mv.grib_get_long(u_2levs, 'level') == [700, 400])


def test_write():
    gg = mv.read({'SOURCE': file_in_testdir('test.grib'), 'GRID': 80})
    regridded_grib = mv.write(file_in_testdir('test_gg_grid.grib'), gg)
    assert regridded_grib == 0
    os.remove(file_in_testdir('test_gg_grid.grib'))


def test_class_():
    # these generate warnings, but if they pass then they show that the conversion
    # from class_ to class is working
    gg = mv.read(file_in_testdir('test.grib'))
    c = mv.read(data=gg, class_='od')
    assert(mv.type(c) == 'fieldset')
    c = mv.read({'data': gg, 'class_' : 'od'})
    assert(mv.type(c) == 'fieldset')
    c = mv.read({'data': gg, 'class' : 'od'})
    assert(mv.type(c) == 'fieldset')


TEST_FIELDSET = mv.read(os.path.join(PATH, 'test.grib'))


def test_type():
    out = mv.type(TEST_FIELDSET)
    assert out == 'fieldset'


# def test_retrieve():
#     tccp = mv.retrieve({
#         'levtype': 'sfc',
#         'param': 'tccp',
#         'grid': 'o640'  # octahedral grid (a specific form of a reduced Gaussian grid)
#     })
#     assert mv.type(tccp)== 'fieldset'


def test_count():
    out = mv.count(TEST_FIELDSET)
    assert out == 1


def test_maxvalue():
    maximum = mv.maxvalue(TEST_FIELDSET)
    assert np.isclose(maximum, MAX_VALUE)


def test_accumulate():
    all_missing = mv.read(file_in_testdir('all_missing_vals.grib'))
    out = mv.accumulate(all_missing)
    assert out is None


def test_add():
    plus_two = TEST_FIELDSET + 2
    maximum = mv.maxvalue(plus_two)
    assert np.isclose(maximum, MAX_VALUE + 2)


def test_add_fieldsets():
    sum = TEST_FIELDSET + TEST_FIELDSET
    maximum = mv.maxvalue(sum)
    assert np.isclose(maximum, MAX_VALUE + MAX_VALUE)


def test_sub():
    minus_two = TEST_FIELDSET - 2
    maximum = mv.maxvalue(minus_two)
    assert np.isclose(maximum, MAX_VALUE - 2)


def test_sub_fieldsets():
    sub = TEST_FIELDSET - TEST_FIELDSET
    maximum = mv.maxvalue(sub)
    assert np.isclose(maximum, 0)


def test_sqrt():
    sqrt_fd = mv.sqrt(TEST_FIELDSET)
    maximum = mv.maxvalue(sqrt_fd)
    assert np.isclose(maximum, np.sqrt(MAX_VALUE))


def test_product():
    times_two = TEST_FIELDSET * 2
    maximum = mv.maxvalue(times_two)
    assert np.isclose(maximum, MAX_VALUE * 2)


def test_product_fieldsets():
    prod = TEST_FIELDSET * TEST_FIELDSET
    maximum = mv.maxvalue(prod)
    assert np.isclose(maximum, MAX_VALUE * MAX_VALUE)


def test_division():
    divided_two = TEST_FIELDSET / 2
    maximum = mv.maxvalue(divided_two)
    assert np.isclose(maximum, MAX_VALUE / 2)


def test_division_fieldsets():
    div = TEST_FIELDSET / TEST_FIELDSET
    maximum = mv.maxvalue(div)
    assert np.isclose(maximum, 1)


def test_power():
    raised_two = TEST_FIELDSET ** 2
    maximum = mv.maxvalue(raised_two)
    assert np.isclose(maximum, MAX_VALUE ** 2)


def test_distance():
    dist = mv.distance(TEST_FIELDSET, 0, 0)
    minimum = mv.minvalue(dist)
    maximum = mv.maxvalue(dist)
    assert np.isclose(minimum, 0.0)
    assert np.isclose(maximum, SEMI_EQUATOR)


def test_valid_date_len_1():
    vd = mv.valid_date(TEST_FIELDSET)
    assert isinstance(vd, datetime.datetime)
    assert vd == datetime.datetime(2017, 4, 27, 12, 0, 0)


def test_valid_date_len_6():
    grib = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    vd_grib = mv.valid_date(grib)
    assert isinstance(vd_grib[1], datetime.datetime)

    vd_ref = datetime.datetime(2017, 8, 1, 12, 0, 0)
    for vd in vd_grib:
        assert vd == vd_ref


def test_base_date():
    bd = mv.base_date(TEST_FIELDSET)
    assert isinstance(bd, datetime.datetime)
    assert bd == datetime.datetime(2017, 4, 27, 12, 0, 0)


def test_fieldset_len_1():
    flen = len(TEST_FIELDSET)
    assert(flen == 1)


def test_fieldset_len_6():
    grib = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    flen = len(grib)
    assert(flen == 6)


def test_fieldset_single_index():
    grib = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    grib4 = grib[3]  # 0-based indexing in Python
    assert(len(grib4) == 1)
    assert(mv.grib_get_long(grib4, 'level') == 500)


def test_fieldset_slice():
    grib = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    # slice [start,stop+1,step]
    assert(mv.grib_get_long(grib[0:6:1], 'level') == [1000, 850, 700, 500, 400, 300])
    assert(mv.grib_get_long(grib[0:6:2], 'level') == [1000, 700, 400])
    assert(mv.grib_get_long(grib[4:6:1], 'level') == [400, 300])
    assert(mv.grib_get_long(grib[5:6:1], 'level') == 300)
    assert(mv.grib_get_long(grib[4:12:1], 'level') == [400, 300])
    assert(mv.grib_get_long(grib[1:6:3], 'level') == [850, 400])
    assert(mv.grib_get_long(grib[-1::], 'level') == 300)
    assert(mv.grib_get_long(grib[-1:-4:-1], 'level') == [300, 400, 500])

    grib = mv.read(os.path.join(PATH, 'tuv_pl.grib'))
    assert(mv.grib_get_long(grib[0:18:1], 'level') == [1000, 1000, 1000, 850, 850, 850,
                                                       700, 700, 700, 500, 500, 500,
                                                       400, 400, 400, 300, 300, 300])


def test_fieldset_iterator():
    grib = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    avg = mv.average(grib)
    assert(len(avg) == 6)
    iteravg = []
    for f in grib:
        iteravg.append(mv.average(f))
    assert(len(iteravg) == len(avg))
    for i in range(0, 6):
        assert np.isclose(avg[i], iteravg[i])


def test_fieldset_assignment_to_field_index():
    grib = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    # check numbers before assignment
    assert(np.isclose(mv.integrate(grib[0]), 290.802))
    assert(np.isclose(mv.integrate(grib[3]), 260.601))
    assert(np.isclose(mv.integrate(grib[4]), 249.617))
    # change one field and check some fields
    grib[0] = grib[0] * 10
    assert(np.isclose(mv.integrate(grib[0]), 2908.02))
    assert(np.isclose(mv.integrate(grib[3]), 260.601))
    assert(np.isclose(mv.integrate(grib[4]), 249.617))
    # another assignment
    grib[4] = grib[3] * 10
    assert(np.isclose(mv.integrate(grib[0]), 2908.02))
    assert(np.isclose(mv.integrate(grib[3]), 260.601))
    assert(np.isclose(mv.integrate(grib[4]), 2606.01))


def test_fieldset_relational_operators():
    a = mv.read(os.path.join(PATH, 'test.grib'))
    a = mv.int(a)
    assert(mv.accumulate(a > 273) == 76001)
    assert(mv.accumulate(a >= 273) == 78156)
    assert(mv.accumulate(a < 273) == 37524)
    assert(mv.accumulate(a <= 273) == 39679)


def test_fieldset_equality_operator():
    a = mv.read(os.path.join(PATH, 'test.grib'))
    v = mv.values(a)
    v[10] = -29
    v[13] = -31
    v[15] = -33
    b = a.set_values(v)
    same = (a == b)
    assert(type(same) == mv.Fieldset)
    vsame = same.values()
    assert(mv.sum(vsame) == len(vsame) - 3)  # all but 3 should be the same
    assert(vsame[5]  == 1)
    assert(vsame[10] == 0)
    assert(vsame[11] == 1)
    assert(vsame[12] == 1)
    assert(vsame[13] == 0)
    assert(vsame[14] == 1)
    assert(vsame[15] == 0)


def test_fieldset_nonequality_operator():
    a = mv.read(os.path.join(PATH, 'test.grib'))
    v = mv.values(a)
    v[10] = -29
    v[13] = -31
    v[15] = -33
    b = a.set_values(v)
    diff = (a != b)
    assert(type(diff) == mv.Fieldset)
    vdiff = diff.values()
    assert(mv.sum(vdiff) == 3)  # there should be exactly 3 differences
    assert(vdiff[5]  == 0)
    assert(vdiff[10] == 1)
    assert(vdiff[11] == 0)
    assert(vdiff[12] == 0)
    assert(vdiff[13] == 1)
    assert(vdiff[14] == 0)
    assert(vdiff[15] == 1)


def test_nearest_gridpoint_info():
    a = mv.read(os.path.join(PATH, 'test.grib'))
    ni = mv.nearest_gridpoint_info(a, 57.193, -2.360)
    ni0 = ni[0]
    assert(np.isclose(ni0['latitude'],  57.0))
    assert(np.isclose(ni0['longitude'], 357.75))
    assert(np.isclose(ni0['distance'],  22.4505))
    assert(np.isclose(ni0['value'],     282.436))
    assert(ni0['index'] == 21597)


def test_surrounding_gridpoints():
    a = mv.read(os.path.join(PATH, 'test.grib'))
    sgi = mv.surrounding_points_indexes(a, 11.552, 50.653)
    assert(np.array_equal(sgi, np.array([50468, 50467, 49988, 49987])))


def test_datainfo():
    a = mv.read(os.path.join(PATH, 'tuv_pl.grib'))
    di = mv.datainfo(a)
    di3 = di[3]
    assert(di3['index'] == 3)
    assert(di3['number_present'] == 2664)
    assert(di3['number_missing'] == 0)
    assert(di3['proportion_present'] == 1)
    assert(di3['proportion_missing'] == 0)


def test_empty_fieldset_contructor():
    f = mv.Fieldset()
    assert(type(f) == mv.Fieldset)
    assert(len(f) == 0)


def test_fieldset_contructor_with_path():
    f = mv.Fieldset(path=os.path.join(PATH, 'test.grib'))
    assert(type(f) == mv.Fieldset)
    assert(len(f) == 1)
    ni = mv.nearest_gridpoint_info(f, 57.193, -2.360)
    ni0 = ni[0]
    assert(np.isclose(ni0['latitude'],  57.0))
    assert(np.isclose(ni0['longitude'], 357.75))
    assert(np.isclose(ni0['distance'],  22.4505))
    assert(np.isclose(ni0['value'],     282.436))
    assert(ni0['index'] == 21597)


def test_fieldset_append_from_empty():
    f = mv.Fieldset()
    g = mv.read(os.path.join(PATH, 'tuv_pl.grib'))
    f.append(g[0])
    assert(type(f) == mv.Fieldset)
    assert(len(f) == 1)
    shortnames = f.grib_get_string('shortName')
    assert(shortnames == 't')
    f.append(g[1])
    assert(type(f) == mv.Fieldset)
    assert(len(f) == 2)
    shortnames = f.grib_get_string('shortName')
    assert(shortnames == ['t', 'u'])
    f.append(g[15:18])
    assert(type(f) == mv.Fieldset)
    assert(len(f) == 5)
    shortnames = f.grib_get_string('shortName')
    assert(shortnames == ['t', 'u', 't', 'u', 'v'])


def test_read_bufr():
    bufr = mv.read(file_in_testdir('obs_3day.bufr'))
    assert(mv.type(bufr) == 'observations')


def test_read_gpt():
    gpt = mv.read(file_in_testdir('t2m_3day.gpt'))
    assert(mv.type(gpt) == 'geopoints')
    assert(mv.count(gpt) == 45)


TEST_GEOPOINTS = mv.read(os.path.join(PATH, 't2m_3day.gpt'))


def test_filter_gpt():
    filter_out = TEST_GEOPOINTS.filter(TEST_GEOPOINTS >= 280)
    assert mv.type(filter_out) == 'geopoints'
    assert mv.count(filter_out) == 38


def test_sqrt_geopoints():
    sqrt_out = mv.sqrt(TEST_GEOPOINTS)
    maximum = mv.maxvalue(sqrt_out)
    assert mv.type(sqrt_out) == 'geopoints'
    assert np.isclose(maximum, MAX_SQRT_GPT)


def test_add_geopoints():
    add = TEST_GEOPOINTS + TEST_GEOPOINTS
    maximum = mv.maxvalue(add)
    assert np.isclose(maximum, MAX_GPT + MAX_GPT)


def test_prod_geopoints():
    prod = TEST_GEOPOINTS * TEST_GEOPOINTS
    maximum = mv.maxvalue(prod)
    assert np.isclose(maximum, MAX_GPT * MAX_GPT)


def test_geopoints_element():
    g1 = TEST_GEOPOINTS[0]
    assert(isinstance(g1, dict))
    assert(g1['latitude']       == 49.43)
    assert(g1['longitude']      == -2.6)
    assert(g1['height']         == 0)
    assert(g1['date']           == 20170425)
    assert(g1['time']           == 1200)
    assert(g1['value']          == 282.4)
    assert(g1['value2']         == 0)
    assert(g1['value_missing']  == 0)
    assert(g1['value2_missing'] == 0)
    g44 = TEST_GEOPOINTS[44]
    assert(isinstance(g44, dict))
    assert(g44['latitude']      == 59.53)


def test_geopoints_relational_operator():
    lt = TEST_GEOPOINTS < 1
    le = TEST_GEOPOINTS <= 1
    gt = TEST_GEOPOINTS > 100
    ge = TEST_GEOPOINTS >= 100
    assert mv.maxvalue(lt) == 0
    assert mv.maxvalue(le) == 0
    assert mv.maxvalue(gt) == 1
    assert mv.maxvalue(ge) == 1


def test_geopoints_equality_operator():
    a = mv.read(file_in_testdir('t2m_3day.gpt'))
    assert(mv.type(a) == 'geopoints')
    v = mv.values(a)
    v[10] = -29
    v[13] = -31
    v[15] = -33
    b = a.set_values(v)
    same = (a == b)
    assert(mv.type(same) == 'geopoints')
    vsame = same.values()
    assert(mv.sum(vsame) == len(vsame) - 3)  # all but 3 should be the same
    assert(vsame[5]  == 1)
    assert(vsame[10] == 0)
    assert(vsame[11] == 1)
    assert(vsame[12] == 1)
    assert(vsame[13] == 0)
    assert(vsame[14] == 1)
    assert(vsame[15] == 0)


def test_geopoints_nonequality_operator():
    a = mv.read(file_in_testdir('t2m_3day.gpt'))
    assert(mv.type(a) == 'geopoints')
    v = mv.values(a)
    v[10] = -29
    v[13] = -31
    v[15] = -33
    b = a.set_values(v)
    diff = (a != b)
    assert(mv.type(diff) == 'geopoints')
    vdiff = diff.values()
    assert(mv.sum(vdiff) == 3)  # there should be exactly 3 differences
    assert(vdiff[5]  == 0)
    assert(vdiff[10] == 1)
    assert(vdiff[11] == 0)
    assert(vdiff[12] == 0)
    assert(vdiff[13] == 1)
    assert(vdiff[14] == 0)
    assert(vdiff[15] == 1)


def test_geopoints_fieldset_operator():
    diff = TEST_FIELDSET - TEST_GEOPOINTS
    assert mv.type(diff) == 'geopoints'


def test_obsfilter():
    bufr = mv.read(file_in_testdir('obs_3day.bufr'))

    # test two styles of passing parameters
    gpt1 = mv.obsfilter({'data': bufr, 'parameter': '012004', 'output': "geopoints"})
    gpt2 = mv.obsfilter(data=bufr, parameter='012004', output="geopoints")
    assert(mv.type(gpt1) == 'geopoints')
    assert(mv.count(gpt1) == 45)
    assert(mv.type(gpt2) == 'geopoints')
    assert(mv.count(gpt2) == 45)


def test_read_gptset():
    gpts = mv.read(file_in_testdir('geopointset_1.gpts'))
    assert(mv.type(gpts) == 'geopointset')
    assert(mv.count(gpts) == 6)
    gpt1 = gpts[0]
    assert(mv.type(gpt1) == 'geopoints')
    assert(mv.count(gpt1) == 11)
    assert(mv.metadata(gpt1) is None)
    gpt2 = gpts[1]
    assert(mv.type(gpt2) == 'geopoints')
    assert(mv.count(gpt2) == 1)
    # check the metadata
    md = mv.metadata(gpt2)
    assert(isinstance(md, dict))
    assert(md['mykey1'] == 'val1')
    assert(md['mykey2'] == 5)
    # check that it is iterable
    counts = [mv.count(c) for c in gpts]
    assert(counts == [11.0, 1.0, 44.0, 11.0, 1.0, 44.0])
    # test the filtering
    bad_filter = mv.filter(gpts, {'badkey' : 7})
    assert(bad_filter is None)
    good_filter = mv.filter(gpts, {'mykey2' : 5})
    assert(mv.type(good_filter) == 'geopointset')
    assert(mv.count(good_filter) == 1)
    assert(mv.count(good_filter[0]) == 1)
    lats = good_filter[0].latitudes()
    assert(len(lats) == 1)
    assert(lats[0] == 60.82)


def test_date_year():
    npd1 = np.datetime64("2017-04-27T06:18:02")
    assert mv.year(npd1) == 2017
    npd2 = np.datetime64("2017-04-27T06:18:02.16")
    assert mv.year(npd2) == 2017

    dt1 = datetime.datetime(2017, 4, 27, 6, 18, 2)
    assert mv.year(dt1) == 2017
    dt2 = datetime.datetime(2017, 4, 27, 6, 18, 2, 16)
    assert mv.year(dt2) == 2017

    d1 = datetime.date(2017, 4, 27)
    assert mv.year(d1) == 2017


def test_date_month():
    npd1 = np.datetime64("2017-04-27T06:18:02")
    assert mv.month(npd1) == 4
    npd2 = np.datetime64("2017-04-27T06:18:02.16")
    assert mv.month(npd2) == 4

    dt1 = datetime.datetime(2017, 4, 27, 6, 18, 2)
    assert mv.month(dt1) == 4
    dt2 = datetime.datetime(2017, 4, 27, 6, 18, 2, 16)
    assert mv.month(dt2) == 4

    d1 = datetime.date(2017, 4, 27)
    assert mv.month(d1) == 4


def test_date_day():
    npd1 = np.datetime64("2017-04-27T06:18:02")
    assert mv.day(npd1) == 27
    npd2 = np.datetime64("2017-04-27T06:18:02.16")
    assert mv.day(npd2) == 27

    dt1 = datetime.datetime(2017, 4, 27, 6, 18, 2)
    assert mv.day(dt1) == 27
    dt2 = datetime.datetime(2017, 4, 27, 6, 18, 2, 16)
    assert mv.day(dt2) == 27

    d1 = datetime.date(2017, 4, 27)
    assert mv.day(d1) == 27


def test_date_hour():
    npd1 = np.datetime64("2017-04-27T06:18:02")
    assert mv.hour(npd1) == 6
    npd2 = np.datetime64("2017-04-27T06:18:02.16")
    assert mv.hour(npd2) == 6

    dt1 = datetime.datetime(2017, 4, 27, 6, 18, 2)
    assert mv.hour(dt1) == 6
    dt2 = datetime.datetime(2017, 4, 27, 6, 18, 2, 16)
    assert mv.hour(dt2) == 6

    d1 = datetime.date(2017, 4, 27)
    assert mv.hour(d1) == 0


def test_date_minute():
    npd1 = np.datetime64("2017-04-27T06:18:02")
    assert mv.minute(npd1) == 18
    npd2 = np.datetime64("2017-04-27T06:18:02.16")
    assert mv.minute(npd2) == 18

    dt1 = datetime.datetime(2017, 4, 27, 6, 18, 2)
    assert mv.minute(dt1) == 18
    dt2 = datetime.datetime(2017, 4, 27, 6, 18, 2, 16)
    assert mv.minute(dt2) == 18

    d1 = datetime.date(2017, 4, 27)
    assert mv.minute(d1) == 0


def test_date_second():
    npd1 = np.datetime64("2017-04-27T06:18:02")
    assert mv.second(npd1) == 2
    npd2 = np.datetime64("2017-04-27T06:18:02.16")
    assert mv.second(npd2) == 2

    dt1 = datetime.datetime(2017, 4, 27, 6, 18, 2)
    assert mv.second(dt1) == 2
    dt2 = datetime.datetime(2017, 4, 27, 6, 18, 2, 16)
    assert mv.second(dt2) == 2

    d1 = datetime.date(2017, 4, 27)
    assert mv.second(d1) == 0


def test_numpy_numeric_args():
    # we don't test them all here, but hopefully a representative sample
    assert(mv.call('+', np.int32(5), 1) == 6)
    assert(mv.call('+', np.int64(5), 1) == 6)
    assert(mv.call('+', np.byte(8), 1) == 9)
    assert(mv.call('+', np.uint16(8), 1) == 9)
    assert(np.isclose(mv.call('+', np.float32(5.5), 1), 6.5))
    assert(np.isclose(mv.call('+', np.float64(5.5), 1), 6.5))


def test_odb():
    if mv.is_feature_available('odb') == 0:
        print('Skipping test_odb because ODB is not enabled in this Metview version')
        return

    db = mv.read(file_in_testdir('temp_u.odb'))
    assert(mv.type(db) == 'odb')

    # assert isinstance(db,mv.Odb)
    assert(mv.count(db) == 88)

    p_val = mv.values(db, 'p')
    assert(mv.count(p_val) == 88)
    assert(np.isclose(p_val[0], 98065.578125))
    assert(np.isclose(p_val[87], 97651.2109375))

    t_val = mv.values(db, 't')
    assert(mv.count(t_val) == 88)
    assert(np.isclose(t_val[0], 144700))
    assert(np.isclose(t_val[87], 94700))

    v_val = mv.values(db, 'val')
    assert(mv.count(v_val) == 88)
    assert(np.isclose(v_val[0], -4.62306786))
    assert(np.isclose(v_val[87], -4.27525187))


def test_odb_filter():
    if mv.is_feature_available('odb') == 0:
        print('Skipping test_odb_filter because ODB is not enabled in this Metview version')
        return

    db = mv.read(file_in_testdir('temp_u.odb'))
    # assert isinstance(db,mv.Odb)
    assert(mv.count(db) == 88)

    db_res = mv.odb_filter({
        'odb_data': db,
        'odb_query': "select p, t, val where val < -8",
    })

    # assert isinstance(db_res,mv.Odb)
    assert(mv.count(db_res) == 6)
    val = mv.values(db_res, 'val')
    a = np.array([-9.11442089, -8.16880512, -8.07200909, -8.06602955, -8.49743557, -8.21722794])
    np.testing.assert_allclose(val, a)


def test_odb_to_dataframe_1():
    if mv.is_feature_available('odb') == 0:
        print('Skipping test_odb because ODB is not enabled in this Metview version')
        return

    db = mv.read(file_in_testdir('temp_u.odb'))
    assert(mv.type(db) == 'odb')

    df = db.to_dataframe()
    assert(isinstance(df, pd.DataFrame))

    assert(df.shape == (88, 3))
    dt1_loc = df.iloc[0]['p']
    assert(np.isclose(dt1_loc, 98065.6))
    dt1_loc = df.iloc[5]['t']
    assert(np.isclose(dt1_loc, 144700))
    dt1_loc = df.iloc[50]['t']
    assert(np.isclose(dt1_loc, 103200))
    dt1_loc = df.iloc[29]['val']
    assert(np.isclose(dt1_loc, -6.06863))
    dt1_loc = df.iloc[87]['val']
    assert(np.isclose(dt1_loc, -4.27525))


# test an ODB that has strings
def test_odb_to_dataframe_2():
    if mv.is_feature_available('odb') == 0:
        print('Skipping test_odb because ODB is not enabled in this Metview version')
        return

    db = mv.read(file_in_testdir('small_odb.odb'))
    assert(mv.type(db) == 'odb')

    df = db.to_dataframe()
    assert(isinstance(df, pd.DataFrame))

    assert(df.shape == (273, 54))
    dt1_loc = df.iloc[0]['an_depar@body']
    assert(np.isclose(dt1_loc, -0.123831))
    dt1_loc = df.iloc[0]['class@desc']
    assert(dt1_loc.strip() == 'rd')      # strings from ODB are padded with spaces
    dt1_loc = df.iloc[20]['expver@desc']
    assert(dt1_loc.strip() == 'fgww')    # strings from ODB are padded with spaces
    dt1_loc = df.iloc[20]['sensor@hdr']  # test a constant integer field
    assert(dt1_loc == 3)


# this tests the calling of the Cross Section module, but also the
# return of netCDF data and also that we can perform operations on it
# as input and output
def test_cross_section_data():
    grib = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    xs_data = mv.mcross_sect(
        line=[59.9, -180, -13.5, 158.08],
        data=grib,
    )
    # the result of this should be a netCDF variable
    assert mv.type(xs_data) == 'netcdf'
    mv.setcurrent(xs_data, 't')
    assert mv.dimension_names(xs_data) == ['time', 'nlev', 'lon']
    assert np.isclose(mv.value(xs_data, 0), 230.39156)
    xs_data_x2 = xs_data * 2
    assert np.isclose(mv.value(xs_data_x2, 0), 460.7831)


def test_netcdf_var_indexing():
    nc = mv.read(file_in_testdir('xs_date_mv5.nc'))
    mv.setcurrent(nc, 0)
    assert(mv.attributes(nc)['long_name'] == "time")
    mv.setcurrent(nc, 1)
    assert(mv.attributes(nc)['long_name'] == "latitude")
    mv.setcurrent(nc, 4)
    assert(mv.attributes(nc)['long_name'] == "Temperature")


def test_netcdf_single_value():
    nc = mv.read(file_in_testdir('xs_date_mv5.nc'))
    mv.setcurrent(nc, 't')
    assert(mv.attributes(nc)['long_name'] == "Temperature")
    assert(np.isclose(mv.value(nc, 0), 234.7144))
    assert(np.isclose(mv.value(nc, 4), 237.4377))


def test_netcdf_multi_indexed_values():
    nc = mv.read(file_in_testdir('xs_date_mv5.nc'))
    mv.setcurrent(nc, 't')
    assert(mv.attributes(nc)['long_name'] == "Temperature")
    assert(np.isclose(mv.values(nc, [0, 0, 0]), 234.7144))
    assert(np.isclose(mv.values(nc, [0, 0, 4]), 237.4377))
    assert(np.isclose(mv.values(nc, [0, 1, 0]), 248.7220))
    assert(np.isclose(mv.values(nc, [0, 1, 1]), 249.3030))


def test_netcdf_multi_indexed_values_with_all():
    nc = mv.read(file_in_testdir('xs_date_mv5.nc'))
    mv.setcurrent(nc, 't')
    assert(mv.attributes(nc)['long_name'] == "Temperature")
    v = mv.values(nc, [0, 0, 'all'])
    assert(len(v) == 64)
    assert(np.isclose(v[0],  234.714))
    assert(np.isclose(v[63], 258.979))
    v = mv.values(nc, [0, 'all', 0])
    assert(len(v) == 5)
    assert(np.isclose(v[0], 234.714))
    assert(np.isclose(v[4], 260.484))


def test_netcdf_to_dataset():
    nc = mv.read(file_in_testdir('xs_date_mv5.nc'))
    x = nc.to_dataset()
    assert(isinstance(x, xr.core.dataset.Dataset))
    assert(isinstance(x['t'], xr.DataArray))


@pytest.mark.filterwarnings("ignore:GRIB write")
def test_dataset_to_fieldset():
    grib = mv.read(file_in_testdir('t_for_xs.grib'))
    x = grib.to_dataset()
    f = mv.dataset_to_fieldset(x)
    assert(mv.type(f) == 'fieldset')
    assert(len(f) == 6)


@pytest.mark.filterwarnings("ignore:GRIB write")
def test_pass_dataset_as_arg():
    grib = mv.read(file_in_testdir('t_for_xs.grib'))
    x = grib.to_dataset()
    fs = mv.mean(x * 2)  # *2 is done by xarray, mean is done by Metview
    assert(mv.type(fs) == 'fieldset')
    assert(len(fs) == 1)


def test_strings():
    sentence = "Hello this is a sentence with many many words"
    sthis = mv.search(sentence, "this")
    assert(sthis == 6)


@pytest.mark.skip()
def test_met_plot():
    contour = mv.mcont({
        'CONTOUR_LINE_COLOUR': 'PURPLE',
        'CONTOUR_LINE_THICKNESS': 3,
        'CONTOUR_HIGHLIGHT': False
    })
    coast = mv.mcoast({'MAP_COASTLINE_LAND_SHADE': True})
    bindings.met_plot(TEST_FIELDSET, contour, coast)


def test_plot_1():
    output_name = file_in_testdir('test_plot_1')
    mv.setoutput(mv.png_output(output_name=output_name))
    grid_shade = {
        'legend': True,
        'contour': False,
        'contour_highlight': True,
        'contour_shade': True,
        'contour_shade_technique': 'grid_shading',
        'contour_shade_max_level_colour': 'red',
        'contour_shade_min_level_colour': 'blue',
        'contour_shade_colour_direction': 'clockwise',
    }
    degraded_field = mv.read(data=TEST_FIELDSET, grid=[4, 4])
    mv.plot(degraded_field, mv.mcont(grid_shade))
    output_name_from_magics = output_name + '.1.png'
    assert(os.path.isfile(output_name_from_magics))
    os.remove(output_name_from_magics)


def test_plot_2():
    output_name = file_in_testdir('test_plot_2')
    png = mv.png_output(output_name=output_name)
    grid_shade = {
        'legend': True,
        'contour': False,
        'contour_highlight': True,
        'contour_shade': True,
        'contour_shade_technique': 'grid_shading',
        'contour_shade_max_level_colour': 'olive',
        'contour_shade_min_level_colour': 'blue',
        'contour_shade_colour_direction': 'clockwise',
    }
    degraded_field = mv.read(data=TEST_FIELDSET, grid=[4, 4])
    mv.plot(png, degraded_field, mv.mcont(grid_shade))
    output_name_from_magics = output_name + '.1.png'
    assert(os.path.isfile(output_name_from_magics))
    os.remove(output_name_from_magics)


@pytest.mark.skip()
def test_plot_3():
    png_output = {
        'output_type': 'PnG',
        'output_width': 1200,
        'output_name': file_in_testdir('test_plot')
    }
    grid_shade = {
        'legend': True,
        'contour': False,
        'contour_highlight': True,
        'contour_shade': True,
        'contour_shade_technique': 'grid_shading',
        'contour_shade_max_level_colour': 'red',
        'contour_shade_min_level_colour': 'blue',
        'contour_shade_colour_direction': 'clockwise',
    }
    bindings.plot(TEST_FIELDSET, grid_shade, **png_output)
    os.remove(file_in_testdir('test_plot.1.png'))


def test_plot_2_pages():
    output_name = file_in_testdir('test_plot_2_pages')
    png = mv.png_output(output_name=output_name)
    degraded_field = mv.read(data=TEST_FIELDSET, grid=[4, 4])
    page1 = mv.plot_page(top=2.2, bottom=52.2, right=100)
    page2 = mv.plot_page(top=50)
    dw = mv.plot_superpage(pages=[page1, page2])
    mv.setoutput(png)
    mv.plot(dw[0], degraded_field, dw[1], degraded_field + 50)
    output_name_from_magics = output_name + '.1.png'
    assert(os.path.isfile(output_name_from_magics))
    os.remove(output_name_from_magics)


def test_macro_error():
    with pytest.raises(Exception):
        TEST_FIELDSET[125]


def test_value_file_path():
    p = TEST_FIELDSET + 1  # this will force Metview to write a new temporary file
    assert(p.url() != "")
    assert(os.path.isfile(p.url()))


@pytest.mark.parametrize('file_name', [
    'ml_data.grib',
    't2m_3day.gpt',
])
def test_temporary_file_deletion(file_name):
    g = mv.read(file_in_testdir(file_name))
    h = g + 1  # this will force Metview to write a new temporary file
    temp_filepath = h.url()
    assert(temp_filepath != "")  # file should exist right now
    assert(os.path.isfile(temp_filepath))  # file should exist right now
    h = 0  # this should force deletion of the variable
    # here we make the assumption that the system has not created
    # another temporary file with the same name between object
    # deletion and the following test for the file
    assert(not(os.path.isfile(temp_filepath)))


def test_mvl_ml2hPa():
    ml_data = mv.read(file_in_testdir('ml_data.grib'))
    assert mv.type(ml_data) == 'fieldset'
    ml_t = mv.read(data=ml_data, param='t')
    ml_lnsp = mv.read(data=ml_data, param='lnsp')
    desired_pls = [1000, 900, 850, 500, 300, 100, 10, 1, 0.8, 0.5, 0.3, 0.1]
    pl_data = mv.mvl_ml2hPa(ml_lnsp, ml_t, desired_pls)
    assert mv.type(pl_data) == 'fieldset'
    pls = mv.grib_get_long(pl_data, 'level')
    lev_types = mv.grib_get_string(pl_data, 'typeOfLevel')
    lev_divisors = [1 if x == 'isobaricInhPa' else 100 for x in lev_types]
    pl_in_hpa = [a / b for a, b in zip(pls, lev_divisors)]
    assert(pl_in_hpa == desired_pls)


def test_push_nil():
    n = mv.nil()
    assert(n is None)
    assert(mv.type(n) == 'nil')


def test_simple_get_vector():
    a = [5, 6, 7, 8, 9]
    v = mv.vector(a)
    assert(isinstance(v, np.ndarray))
    assert(len(v) == 5)
    assert(v[1] == 6)
    assert(v.dtype == np.dtype('float64'))


def test_simple_get_vector_float32():
    if supports_float32_vectors():
        mv.vector_set_default_type('float32')
        a = [5, 6, 7, 8, 9]
        v = mv.vector(a)
        assert(isinstance(v, np.ndarray))
        assert(len(v) == 5)
        assert(v[1] == 6)
        assert(v.dtype == np.dtype('float32'))
        mv.vector_set_default_type('float64')  # reset to default type for the other tests


def test_get_vector_from_grib():
    v = mv.values(TEST_FIELDSET[0])
    assert(isinstance(v, np.ndarray))
    assert(v.dtype == np.dtype('float64'))
    assert(len(v) == 115680)
    assert(np.isclose(min(v), 206.93560791))
    assert(np.isclose(max(v), 316.06060791))


def test_get_vector_float32_from_grib():
    if supports_float32_vectors():
        mv.vector_set_default_type('float32')
        v = mv.values(TEST_FIELDSET[0])
        assert(isinstance(v, np.ndarray))
        assert(v.dtype == np.dtype('float32'))
        assert(len(v) == 115680)
        assert(np.isclose(min(v), 206.93560791))
        assert(np.isclose(max(v), 316.06060791))
        mv.vector_set_default_type('float64')  # reset to default type for the other tests


def test_get_vector_from_multi_field_grib():
    g = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    v = mv.values(g)
    assert(isinstance(v, np.ndarray))
    assert(v.shape == (6, 2664))


def test_vector_sort():
    v1 = np.array([5, 3, 4, 9, 1, 4.2])
    vsortasc = mv.sort(v1)
    assert(np.array_equal(vsortasc, np.array([1, 3, 4, 4.2, 5, 9])))


def test_vector_sort_indices():
    v1 = np.array([5, 3, 4, 9, 1, 4.2])
    vsortindasc = mv.sort_indices(v1)
    assert(np.array_equal(vsortindasc, np.array([4, 1, 2, 5, 0, 3])))


def test_vector_find():
    s = np.array([4, 7, 2, 3, 4, 9, 0], dtype=np.float64)
    f = mv.find(s, 2)
    assert(f == 2)
    f = mv.find(s, 4)
    assert(f == 0)
    f = mv.find(s, 4, 'all')
    assert(np.array_equal(f, np.array([0, 4], dtype=np.float64)))


def test_set_vector_from_numpy_array():
    r = np.arange(1, 21, dtype=np.float64)
    assert(mv.type(r) == 'vector')
    assert(mv.dtype(r) == 'float64')
    assert(mv.count(r) == 20)
    assert(mv.maxvalue(r) == 20)


def test_set_vector_float32_from_numpy_array():
    if supports_float32_vectors():
        r = np.arange(1, 21, dtype=np.float32)
        assert(mv.type(r) == 'vector')
        assert(mv.dtype(r) == 'float32')
        assert(mv.count(r) == 20)
        assert(mv.maxvalue(r) == 20)


def test_simple_vector_with_nans():
    a = np.array([1, np.nan, 2, 3])
    n = -a
    assert(mv.count(a) == 4)
    assert(mv.sum(a) == 6)  # missing vals Python->Macro
    # seems that comparing numPy arrays that have NaNs in them is not
    # so straightforward, but this works:
    np.testing.assert_array_equal(mv.neg(a), n)  # missing vals Macro->Python


def test_zero_length_vector_from_numpy():
    a = np.array([])
    assert(mv.type(a) == 'vector')
    assert(mv.count(a) == 0)


def test_numpy_array_from_zero_length_vector():
    v = mv.vector(0)
    assert(isinstance(v, np.ndarray))
    assert(len(v) == 0)


def test_oo_interface_on_fieldsets():
    fs = mv.read(os.path.join(PATH, 't_for_xs.grib'))
    assert(np.isclose(fs.maxvalue(), 320.434))
    assert(np.isclose(fs[2].nearest_gridpoint(10, 20), 282.697))


def test_oo_interface_on_geopoints():
    gpt = mv.read(file_in_testdir('t2m_3day.gpt'))
    assert(gpt.count() == 45)
    assert(np.isclose(gpt.mean(), 281.247))


def test_std_gpts_to_dataframe():
    gpt = mv.read(file_in_testdir('t2m_3day.gpt'))
    df = gpt.to_dataframe()
    print(df)
    assert(isinstance(df, pd.DataFrame))
    assert(df.shape == (45, 5))
    # dt1_iloc = df.iloc[0][0]
    dt1_iloc = df.iloc[0]['latitude']
    assert(isinstance(dt1_iloc, float))
    assert(np.isclose(dt1_iloc, 49.43))
    # dt2_iloc = df.iloc[0][4]
    dt2_iloc = df.iloc[0]['date']
    assert(isinstance(dt2_iloc, datetime.datetime))
    dt3_loc = df.iloc[0]['date']
    assert(isinstance(dt3_loc, datetime.datetime))
    assert(dt3_loc == datetime.datetime(2017, 4, 25, 12, 0, 0))
    assert(df.loc[5]['latitude'] == 51.15)
    assert(df.loc[25]['longitude'] == 2.65)
    assert(df.loc[8]['level'] == 0)
    assert(np.isclose(df.loc[4]['value'], 281.2))


def test_xy_vector_gpts_to_dataframe():
    gpt = mv.read(file_in_testdir('uv.gpt'))
    df = gpt.to_dataframe()
    assert(isinstance(df, pd.DataFrame))
    assert(df.shape == (39, 6))
    # dt1_iloc = df.iloc[0][0]
    dt1_iloc = df.iloc[0]['latitude']
    assert(isinstance(dt1_iloc, float))
    assert(np.isclose(dt1_iloc, 70.0))
    # dt2_iloc = df.iloc[0][4]
    # assert(isinstance(dt2_iloc, datetime.datetime))
    dt3_loc = df.iloc[0]['date']
    assert(isinstance(dt3_loc, datetime.datetime))
    assert(dt3_loc == datetime.datetime(2018, 5, 14, 12, 0, 0))
    assert(df.loc[5]['latitude'] == 70)
    assert(df.loc[25]['longitude'] == 30)
    assert(df.loc[8]['level'] == 500)
    assert(np.isclose(df.loc[4]['x-comp'], -10.865656))
    assert(np.isclose(df.loc[4]['y-comp'], 17.589124))


def test_xyv_gpts_to_dataframe():
    gpt = mv.read(file_in_testdir('xyv.gpt'))
    df = gpt.to_dataframe()
    assert(isinstance(df, pd.DataFrame))
    assert(df.shape == (60, 3))
    assert(df.loc[5]['latitude'] == 70)
    assert(df.loc[25]['longitude'] == 20)
    assert(np.isclose(df.loc[4]['value'], -10.8656))


def test_grib_to_dataset():
    grib = mv.read(file_in_testdir('t_for_xs.grib'))
    x = grib.to_dataset()
    assert(isinstance(x, xr.core.dataset.Dataset))
    assert(isinstance(x['t'], xr.DataArray))


def test_read_filter_to_dataset():
    tuv_all_levs = mv.read(file_in_testdir('tuv_pl.grib'))
    u_all_levs = mv.read(data=tuv_all_levs, param='u')
    assert(mv.grib_get_long(u_all_levs, 'level') == [1000, 850, 700, 500, 400, 300])
    assert(mv.grib_get_string(u_all_levs, 'shortName') == ['u', 'u', 'u', 'u', 'u', 'u'])
    x = u_all_levs.to_dataset()
    x_keys = x.keys()
    assert('u' in x_keys)      # only 'u' should be there
    assert('v' not in x_keys)  # only 'u' should be there
    assert('t' not in x_keys)  # only 'u' should be there


@pytest.mark.xfail()
def test_table():
    # test csv with metadata
    db = mv.read_table(
        table_filename=file_in_testdir('sample_metadata.csv'),
        table_delimiter=' ',
        table_combine_delimiters='on',
        table_header_row=2,
        table_meta_data_rows=1
    )
    assert(db.type() == 'table')
    assert(db.count() == 9)
    assert(db.name(3) == "LATIT")
    assert(len(db.metadata_keys()) == 16)
    assert(db.metadata_value('integration') == 'PETTERSSEN')
    v = db.values(0)
    assert(isinstance(v, np.ndarray))
    assert(len(v) == 7)
    assert(np.array_equal(v, np.array([0, 3600, 7200, 10800, 14400, 18000, 21600])))
    v = db.values(4)
    assert(isinstance(v, np.ndarray))
    assert(len(v) == 7)
    assert(np.array_equal(v, np.array([963.5, 964.1, 964.0, 963.3, 961.8, 960.3, 959.0])))

    # test csv with no metadata
    db = mv.read(file_in_testdir('sample.csv'))
    assert(db.type() == 'table')
    assert(db.count() == 4)
    assert(db.name(2) == "h2")
    v = db.values(2)
    assert(isinstance(v, np.ndarray))
    assert(len(v) == 6)
    assert(np.array_equal(v, np.array([4, 5, 6, 7, 8, 9])))


def test_flextra():
    flx = mv.read(file_in_testdir('flextra_output.txt'))
    assert(flx.type() == 'definition')
    trNum = int(mv.flextra_group_get(flx, "trNum"))
    assert(trNum == 5)

    startLst = ['03:00:00', '06:00:00', '09:00:00', '12:00:00', '15:00:00']
    stopIndexLst = [1, 1, 1, 1, 1]
    for i in range(trNum):
        vals = mv.flextra_tr_get(flx, i, ["startTime", "stopIndex"])
        assert(len(vals) == 2)
        assert(vals[0] == startLst[i])
        assert(int(vals[1]) == stopIndexLst[i])

    # Read data for the first trajectory
    vals = mv.flextra_tr_get(flx, 0, ["lat", "lon", "date"])
    assert(len(vals) == 3)
    assert(vals[0].size == 25)
    assert(vals[1].size == 25)
    assert(len(vals[2]) == 25)
    assert(isinstance(vals[0], np.ndarray))
    assert(isinstance(vals[1], np.ndarray))
    assert(isinstance(vals[2], list))
    assert(np.isclose(vals[0].mean(), 63.369828))
    assert(np.isclose(vals[0].std(), 2.2269590))
    assert(np.isclose(vals[1].mean(), 18.014544))
    assert(np.isclose(vals[1].std(), 14.98830786689625))
    assert(vals[2][0] == datetime.datetime(2012, 1, 11, 3, 0, 0))



#
# Copyright 2017-2019 B-Open Solutions srl.
# Copyright 2017-2019 European Centre for Medium-Range Weather Forecasts (ECMWF).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import sys


def main(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument('command')
    args = parser.parse_args(args=argv)
    if args.command == 'selfcheck':
        sys.argv = []
        print('Trying to connect to a Metview installation...')
        try:
            from . import bindings as _bindings
        except Exception as exp:
            print('Could not find a valid Metview installation')
            raise(exp)
        mv = dict()
        _bindings.bind_functions(mv, module_name='mv')
        del _bindings

        try:
            mv['print']('Hello world - printed from Metview!')
        except Exception as exp:
            print('Could not print a greeting from Metview')
            raise(exp)

        mv_version_f = mv['version_info']
        mv_version = mv_version_f()
        mv_maj = str(int(mv_version['metview_major']))
        mv_min = str(int(mv_version['metview_minor']))
        mv_rev = str(int(mv_version['metview_revision']))
        mv_version_string = mv_maj + '.' + mv_min + '.' + mv_rev
        print('Metview version', mv_version_string, 'found')

        print("Your system is ready.")
    else:
        raise RuntimeError("Command not recognised %r. See usage with --help." % args.command)


if __name__ == '__main__':
    main()



"""
Metview Python use case

The Python analyst reads some BUFR data and plots it in various ways

--------------------------------------------------------------------------------
1. Python analyst reads BUFR data and plots it using the default style
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
2. Python analyst reads BUFR data and applies a visual definition
   to alter its plotting style
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
3. Python analyst reads BUFR data and filters a single parameter from it
   and plots it with a colour scale
--------------------------------------------------------------------------------

"""
import metview as mv


# define a view over the area of interest and set land shading on

land_shade = mv.mcoast(
    map_coastline_land_shade = True,
    map_coastline_land_shade_colour = "RGB(0.98,0.95,0.82)",
    map_coastline_sea_shade = False,
    map_coastline_sea_shade_colour = "RGB(0.85,0.93,1)"
)

area_view = mv.geoview(
    map_area_definition = 'corners',
    area = [45.83,-13.87,62.03,8.92],
    coastlines = land_shade
)


# Simplest plot:
# NOTE that when plotting a 'raw' BUFR file, Magics will plot synop symbols as shown in
# https://software.ecmwf.int/wiki/display/METV/Data+Part+1 "Plotting BUFR Data"

obs = mv.read('../tests/obs_3day.bufr')

mv.setoutput(mv.png_output(output_width = 1200, output_name = './obsplot1'))
mv.plot(area_view, obs)


# ALTERNATIVELY, add an Observations Plotting visual definition

obs_plotting = mv.mobs(
    obs_temperature = False,
    obs_cloud = False,
    obs_low_cloud = False,
    obs_dewpoint_colour = 'purple'
)


mv.setoutput(mv.png_output(output_width = 1200, output_name = './obsplot2'))
mv.plot(area_view, obs, obs_plotting)


# ALTERNATIVELY, if we don't want to plot the whole observations, but instead want to
# extract a specific parameter from the BUFR messages, then we use the Observation Filter
# as shown here:

# dewpoint_t is a 'geopoints' variable
dewpoint_t = mv.obsfilter(
    output = "geopoints",
    parameter = '012006',
    data = obs
)

# add an optional Symbol Plotting definition to get nice coloured circles
# at each point
symb_visdef = mv.msymb(
    legend = True,
    symbol_type = 'marker',
    symbol_table_mode = 'advanced',
    symbol_advanced_table_max_level_colour = 'red',
    symbol_advanced_table_min_level_colour = 'blue',
    symbol_advanced_table_colour_direction = 'clockwise'
)

mv.setoutput(mv.png_output(output_width = 1200, output_name = './obsplot3'))
mv.plot(area_view, dewpoint_t, symb_visdef)



"""
Metview Python use case

UC-07. The Analyst compute simple differences between observations and analysis
and plot the values

BUFR version - BUFR is not tabular or gridded, but we can use Metview Python
framework to extract a particular parameter to a tabular format (geopoints)

--------------------------------------------------------------------------------
1. Analyst retrieves the analysis from a gridded data file
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
2. Analyst retrieves an observational parameter from a tabular or a gridded file
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
3. Analyst calculates the difference between the observational data and the
   analysis and classified the field values according to the magnitude of the
   difference
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
4. Analyst customises many features of his graph in order to create
   publication-quality plots
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
5. Analyst plots the data
--------------------------------------------------------------------------------
"""

import metview as mv

# define a view over the area of interest

area_view = mv.geoview(
    map_area_definition = 'corners',
    area = [45.83,-13.87,62.03,8.92]
)

t2m_grib = mv.read('./t2m_grib.grib')

obs_3day = mv.read('./obs_3day.bufr')

t2m_gpt = mv.obsfilter(
    parameter = '012004',
    output = 'geopoints',
    data = obs_3day
)

diff = t2m_grib - t2m_gpt

diff_symb = mv.msymb(
    legend = True,
    symbol_type = 'marker',
    symbol_table_mode = 'advanced',
)

mv.setoutput(mv.png_output(output_width = 1000, output_name = './obsdiff1'))
mv.plot(area_view, diff, diff_symb)


# Extract geopoints that are hotter by 1 deg or more
#hotter = mv.filter(diff, diff >= 1)
hotter = diff.filter(diff >= 1)

# Extract geopoints that are colder by 1 deg or more
#colder = mv.filter(diff, diff <= -1)
colder = diff.filter(diff <= -1)

# Get geopoints that are within +/-1
#exact = mv.filter(diff, (diff > -1) * (diff < 1))
exact = diff.filter((diff > -1) * (diff < 1))

# Symbol visdefs for each classification
red = mv.msymb(
    symbol_type = 'marker',
    symbol_colour = 'red'
)

blue  = mv.msymb(
    symbol_type = 'marker',
    symbol_colour = 'blue'
)

grey  = mv.msymb(
    symbol_type = 'marker',
    symbol_colour = 'grey'
)

# plot the 'exact' data set with visdef 'grey', 'hotter' with 'red', etc.
mv.setoutput(mv.png_output(output_width = 1000, output_name = './obsdiff2'))
mv.plot(area_view, exact, grey, hotter, red, colder, blue)



#!/usr/bin/env python3
#
# Copyright 2017-2019 B-Open Solutions srl.
# Copyright 2017-2019 European Centre for Medium-Range Weather Forecasts (ECMWF).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os

import setuptools


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return io.open(file_path, encoding='utf-8').read()


version = '1.1.0'


setuptools.setup(
    name='metview',
    version=version,
    description='Metview Python API.',
    long_description=read('README.rst'),
    author='European Centre for Medium-Range Weather Forecasts (ECMWF)',
    author_email='software.support@ecmwf.int',
    license='Apache License Version 2.0',
    url='https://github.com/ecmwf/metview-python',
    packages=setuptools.find_packages(),
    include_package_data=True,
    setup_requires=[
        'pytest-runner',
    ],
    install_requires=[
        'cffi',
        'numpy',
        'pandas',
    ],
    tests_require=[
        'pytest',
    ],
    test_suite='tests',
    zip_safe=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
    ],
)




import os

from metview import bindings


PATH = os.path.dirname(__file__)
MAX_VALUE = 316.06060791015625
SEMI_EQUATOR = 20001600.0
MAX_SQRT_GPT = 16.867127793433
MAX_GPT = 284.5


def file_in_testdir(filename):
    return os.path.join(PATH, filename)


def test_push_number():
    bindings.lib.p_push_number(5)
    bindings.lib.p_push_number(4)


def test_dict_to_pushed_request():
    dict = {
        'param1': True,
        'param2': False,
        'param3': 10,
        'param4': 10.5,
        'param5': 'metview',
        'param6': ['1', '2', '3']
    }
    bindings.dict_to_pushed_args(dict)


def test_bind_functions():
    namespace = {}

    bindings.bind_functions(namespace, module_name='metview')
    result = namespace['dictionary']

    assert 'dictionary' in namespace
    assert result.__name__ == result.__qualname__ == 'dictionary'
    assert result.__module__ == 'metview'


def test_lists_as_input():
    my_list = [1, 5, 6]
    assert bindings.count(my_list) == 3


def test_tuples_as_input():
    my_tuple = [1, 0, 5, 6]
    assert bindings.count(my_tuple) == 4



"""
Metview Python use case

UC-07-pandas. The Analyst compute simple differences between observations and analysis
and use pandas to perform further computations

BUFR version - BUFR is not tabular or gridded, but we can use Metview Python
framework to extract a particular parameter to a tabular format (geopoints)

--------------------------------------------------------------------------------
1. Analyst retrieves the analysis from a gridded data file
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
2. Analyst retrieves an observational parameter from a tabular or a gridded file
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
3. Analyst calculates the difference between the observational data and the
   analysis
--------------------------------------------------------------------------------

--------------------------------------------------------------------------------
4. Analyst converts this data to a pandas dataframe and computes the number
   of outliers based on the zscore
--------------------------------------------------------------------------------
"""

import metview as mv
import numpy as np
from scipy import stats

t2m_grib = mv.read('./t2m_grib.grib')

obs_3day = mv.read('./obs_3day.bufr')

t2m_gpt = mv.obsfilter(
    parameter = '012004',
    output = 'geopoints',
    data = obs_3day
)

diff = t2m_grib - t2m_gpt

df = diff.to_dataframe()

print(df)

outliers = np.abs(stats.zscore(df['value'])) > 1.5

print('# of outliers:', outliers.sum())



"""
Metview Python use case

UC-01. The Analyst produces plots and files for the Product user

--------------------------------------------------------------------------------
1. Analyst creates plots and files thanks to his Python applications and scripts
   that benefits from the underlying tools of the framework
--------------------------------------------------------------------------------

Analyst reads data from a GRIB file and derives another quantity from it. Then,
Analyst saves his data as a GRIB file and creates a plot in PNG format. 
"""

import metview as mv


mydata = mv.read('../tests/test.grib')

derived = mydata * 2 + 5

mv.write('derived_data.grib', derived)

grid_shade = mv.mcont(
    legend = True,
    contour = False,
    contour_highlight = True,
    contour_shade = True,
    contour_shade_technique = 'grid_shading',
    contour_shade_max_level_colour = 'red',
    contour_shade_min_level_colour = 'blue',
    contour_shade_colour_direction = 'clockwise',
)


# Macro-like PNG creation:
png = mv.png_output(output_width = 1200, output_name = './myplot')

mv.plot(png, derived, grid_shade)


# Using a different notation:
png_output = {
    'output_type': 'png',
    'output_width': 1200,
    'output_name': './myplot2'
}

mv.plot(derived, grid_shade, **png_output)
